import json
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.reports import codes as report_codes
from pcs.lib.dr.config.facade import DrRole
from pcs.lib.commands import dr

DR_CFG_DESC = "disaster-recovery config"

COROSYNC_CONF_TEMPLATE = """\
totem {{
    version: 2
    cluster_name: cluster_name
}}

nodelist {{
{node_list}}}
"""

NODE_TEMPLATE_NO_NAME = """\
    node {{
        ring0_addr: {node}
        nodeid: {id}
    }}
"""

NODE_TEMPLATE = """\
    node {{
        ring0_addr: {node}
        name: {node}
        nodeid: {id}
    }}
"""


def export_cfg(cfg_struct):
    return json.dumps(cfg_struct, indent=4, sort_keys=True).encode("utf-8")


def dr_cfg_fixture(local_role, remote_role, nodes):
    return export_cfg(
        dict(
            local=dict(
                role=local_role.value,
            ),
            remote_sites=[
                dict(
                    role=remote_role.value,
                    nodes=[dict(name=node) for node in nodes],
                ),
            ],
        )
    )


def corosync_conf_fixture(node_list):
    return COROSYNC_CONF_TEMPLATE.format(
        node_list="\n".join(node_list_fixture(node_list)),
    )


def node_list_fixture(node_list):
    return [
        NODE_TEMPLATE.format(node=node, id=i)
        for i, node in enumerate(node_list, start=1)
    ]


def generate_nodes(nodes_num, prefix=""):
    return [f"{prefix}node{i}" for i in range(1, nodes_num + 1)]


class CheckLive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def assert_live_required(self, forbidden_options):
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), "node"),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=forbidden_options,
                )
            ],
            expected_in_processor=False,
        )

    def test_mock_corosync(self):
        self.config.env.set_corosync_conf_data(
            corosync_conf_fixture(generate_nodes(3))
        )
        self.assert_live_required([file_type_codes.COROSYNC_CONF])

    def test_mock_cib(self):
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required([file_type_codes.CIB])

    def test_mock(self):
        self.config.env.set_corosync_conf_data(
            corosync_conf_fixture(generate_nodes(3))
        )
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required(
            [
                file_type_codes.CIB,
                file_type_codes.COROSYNC_CONF,
            ]
        )


class SetRecoverySiteSuccess(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def _test_minimal(self, local_cluster_size, recovery_cluster_size):
        local_nodes = generate_nodes(local_cluster_size)
        remote_nodes = generate_nodes(recovery_cluster_size, prefix="recovery-")
        orig_node = remote_nodes[-1]
        cfg = self.config
        cfg.env.set_known_nodes(local_nodes + remote_nodes)
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(local_nodes))
        cfg.http.corosync.get_corosync_conf(
            corosync_conf_fixture(remote_nodes), node_labels=[orig_node]
        )
        cfg.http.files.put_files(
            node_labels=remote_nodes,
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.RECOVERY, DrRole.PRIMARY, local_nodes
            ),
            name="distribute_remote",
        )
        cfg.http.files.put_files(
            node_labels=local_nodes,
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.PRIMARY, DrRole.RECOVERY, remote_nodes
            ),
            name="distribute_local",
        )
        dr.set_recovery_site(self.env_assist.get_env(), orig_node)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.FILES_DISTRIBUTION_STARTED,
                    file_list=[DR_CFG_DESC],
                    node_list=remote_nodes,
                )
            ]
            + [
                fixture.info(
                    report_codes.FILE_DISTRIBUTION_SUCCESS,
                    file_description=DR_CFG_DESC,
                    node=node,
                )
                for node in remote_nodes
            ]
            + [
                fixture.info(
                    report_codes.FILES_DISTRIBUTION_STARTED,
                    file_list=[DR_CFG_DESC],
                    node_list=local_nodes,
                )
            ]
            + [
                fixture.info(
                    report_codes.FILE_DISTRIBUTION_SUCCESS,
                    file_description=DR_CFG_DESC,
                    node=node,
                )
                for node in local_nodes
            ]
        )

    def test_minimal_local_1_remote_1(self):
        self._test_minimal(1, 1)

    def test_minimal_local_1_remote_2(self):
        self._test_minimal(1, 2)

    def test_minimal_local_1_remote_3(self):
        self._test_minimal(1, 3)

    def test_minimal_local_2_remote_1(self):
        self._test_minimal(2, 1)

    def test_minimal_local_2_remote_2(self):
        self._test_minimal(2, 2)

    def test_minimal_local_2_remote_3(self):
        self._test_minimal(2, 3)

    def test_minimal_local_3_remote_1(self):
        self._test_minimal(3, 1)

    def test_minimal_local_3_remote_2(self):
        self._test_minimal(3, 2)

    def test_minimal_local_3_remote_3(self):
        self._test_minimal(3, 3)


class FailureValidations(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.local_nodes = generate_nodes(4)

    def test_dr_cfg_exist(self):
        orig_node = "node"
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes + [orig_node])
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=True,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(self.local_nodes))
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.DR_CONFIG_ALREADY_EXIST,
                )
            ]
        )

    def test_local_nodes_name_missing(self):
        orig_node = "node"
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes + [orig_node])
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(
            COROSYNC_CONF_TEMPLATE.format(
                node_list="\n".join(
                    [
                        NODE_TEMPLATE_NO_NAME.format(
                            node=self.local_nodes[0], id=len(self.local_nodes)
                        )
                    ]
                    + node_list_fixture(self.local_nodes[1:])
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                )
            ]
        )

    def test_node_part_of_local_cluster(self):
        orig_node = self.local_nodes[-1]
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes + [orig_node])
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(self.local_nodes))
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_IN_LOCAL_CLUSTER,
                    node=orig_node,
                )
            ]
        )

    def test_tokens_missing_for_local_nodes(self):
        orig_node = "node"
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes[:-1] + [orig_node])
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(self.local_nodes))
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    host_list=self.local_nodes[-1:],
                )
            ]
        )

    def test_token_missing_for_node(self):
        orig_node = "node"
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes)
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(self.local_nodes))
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    host_list=[orig_node],
                )
            ]
        )

    def test_tokens_missing_for_remote_cluster(self):
        remote_nodes = generate_nodes(3, prefix="recovery-")
        orig_node = remote_nodes[0]
        cfg = self.config
        cfg.env.set_known_nodes(self.local_nodes + remote_nodes[:-1])
        cfg.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        cfg.corosync_conf.load_content(corosync_conf_fixture(self.local_nodes))
        cfg.http.corosync.get_corosync_conf(
            corosync_conf_fixture(remote_nodes), node_labels=[orig_node]
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), orig_node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    host_list=remote_nodes[-1:],
                )
            ]
        )


REASON = "error msg"


class FailureRemoteCorocyncConf(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.local_nodes = generate_nodes(4)
        self.remote_nodes = generate_nodes(3, prefix="recovery-")
        self.node = self.remote_nodes[0]

        self.config.env.set_known_nodes(self.local_nodes + self.remote_nodes)
        self.config.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.local_nodes)
        )

    def test_network_issue(self):
        self.config.http.corosync.get_corosync_conf(
            communication_list=[
                dict(
                    label=self.node,
                    was_connected=False,
                    error_msg=REASON,
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.node,
                    command="remote/get_corosync_conf",
                    reason=REASON,
                ),
                fixture.error(
                    report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE
                ),
            ]
        )

    def test_file_does_not_exist(self):
        self.config.http.corosync.get_corosync_conf(
            communication_list=[
                dict(
                    label=self.node,
                    response_code=400,
                    output=REASON,
                )
            ]
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.node,
                    command="remote/get_corosync_conf",
                    reason=REASON,
                ),
                fixture.error(
                    report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE
                ),
            ]
        )

    def test_node_names_missing(self):
        self.config.http.corosync.get_corosync_conf(
            COROSYNC_CONF_TEMPLATE.format(
                node_list="\n".join(
                    [
                        NODE_TEMPLATE_NO_NAME.format(
                            node=self.remote_nodes[-1],
                            id=len(self.remote_nodes),
                        )
                    ]
                    + node_list_fixture(self.remote_nodes[:-1])
                )
            ),
            node_labels=[self.node],
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                )
            ]
        )


class FailureRemoteDrCfgDistribution(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.local_nodes = generate_nodes(4)
        self.remote_nodes = generate_nodes(3, prefix="recovery-")
        self.node = self.remote_nodes[0]
        self.failed_nodes = self.remote_nodes[-1:]
        successful_nodes = self.remote_nodes[:-1]

        self.config.env.set_known_nodes(self.local_nodes + self.remote_nodes)
        self.config.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(self.local_nodes)
        )
        self.config.http.corosync.get_corosync_conf(
            corosync_conf_fixture(self.remote_nodes), node_labels=[self.node]
        )

        self.success_communication = [
            dict(label=node) for node in successful_nodes
        ]
        self.expected_reports = [
            fixture.info(
                report_codes.FILES_DISTRIBUTION_STARTED,
                file_list=[DR_CFG_DESC],
                node_list=self.remote_nodes,
            )
        ] + [
            fixture.info(
                report_codes.FILE_DISTRIBUTION_SUCCESS,
                file_description=DR_CFG_DESC,
                node=node,
            )
            for node in successful_nodes
        ]

    def test_write_failure(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    output=json.dumps(
                        dict(
                            files={
                                DR_CFG_DESC: dict(
                                    code="unexpected", message=REASON
                                ),
                            }
                        )
                    ),
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.RECOVERY, DrRole.PRIMARY, self.local_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.FILE_DISTRIBUTION_ERROR,
                    file_description=DR_CFG_DESC,
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )

    def test_network_failure(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    was_connected=False,
                    error_msg=REASON,
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.RECOVERY, DrRole.PRIMARY, self.local_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/put_file",
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )

    def test_communication_error(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    response_code=400,
                    output=REASON,
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.RECOVERY, DrRole.PRIMARY, self.local_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    command="remote/put_file",
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )


class FailureLocalDrCfgDistribution(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        local_nodes = generate_nodes(4)
        self.remote_nodes = generate_nodes(3, prefix="recovery-")
        self.node = self.remote_nodes[0]
        self.failed_nodes = local_nodes[-1:]
        successful_nodes = local_nodes[:-1]

        self.config.env.set_known_nodes(local_nodes + self.remote_nodes)
        self.config.raw_file.exists(
            file_type_codes.PCS_DR_CONFIG,
            settings.pcsd_dr_config_location,
            exists=False,
        )
        self.config.corosync_conf.load_content(
            corosync_conf_fixture(local_nodes)
        )
        self.config.http.corosync.get_corosync_conf(
            corosync_conf_fixture(self.remote_nodes), node_labels=[self.node]
        )
        self.config.http.files.put_files(
            node_labels=self.remote_nodes,
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.RECOVERY, DrRole.PRIMARY, local_nodes
            ),
            name="distribute_remote",
        )

        self.success_communication = [
            dict(label=node) for node in successful_nodes
        ]
        self.expected_reports = (
            [
                fixture.info(
                    report_codes.FILES_DISTRIBUTION_STARTED,
                    file_list=[DR_CFG_DESC],
                    node_list=self.remote_nodes,
                )
            ]
            + [
                fixture.info(
                    report_codes.FILE_DISTRIBUTION_SUCCESS,
                    file_description=DR_CFG_DESC,
                    node=node,
                )
                for node in self.remote_nodes
            ]
            + [
                fixture.info(
                    report_codes.FILES_DISTRIBUTION_STARTED,
                    file_list=[DR_CFG_DESC],
                    node_list=local_nodes,
                )
            ]
            + [
                fixture.info(
                    report_codes.FILE_DISTRIBUTION_SUCCESS,
                    file_description=DR_CFG_DESC,
                    node=node,
                )
                for node in successful_nodes
            ]
        )

    def test_write_failure(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    output=json.dumps(
                        dict(
                            files={
                                DR_CFG_DESC: dict(
                                    code="unexpected", message=REASON
                                ),
                            }
                        )
                    ),
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.PRIMARY, DrRole.RECOVERY, self.remote_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.FILE_DISTRIBUTION_ERROR,
                    file_description=DR_CFG_DESC,
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )

    def test_network_failure(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    was_connected=False,
                    error_msg=REASON,
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.PRIMARY, DrRole.RECOVERY, self.remote_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/put_file",
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )

    def test_communication_error(self):
        self.config.http.files.put_files(
            communication_list=self.success_communication
            + [
                dict(
                    label=node,
                    response_code=400,
                    output=REASON,
                )
                for node in self.failed_nodes
            ],
            pcs_disaster_recovery_conf=dr_cfg_fixture(
                DrRole.PRIMARY, DrRole.RECOVERY, self.remote_nodes
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.set_recovery_site(self.env_assist.get_env(), self.node),
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    command="remote/put_file",
                    reason=REASON,
                    node=node,
                )
                for node in self.failed_nodes
            ]
        )



import os.path
from typing import (
    Iterable,
    List,
    Mapping,
    NamedTuple,
)

from lxml.etree import _Element

from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.node_communicator import Communicator
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.str_tools import (
    format_list,
    indent,
)
from pcs.lib.cib import nvpair, stonith
from pcs.lib.cib.tools import get_crm_config, get_resources
from pcs.lib.communication.nodes import CheckReachability
from pcs.lib.communication.tools import run as run_communication
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.node_communication import NodeTargetLibFactory
from pcs.lib.pacemaker.live import (
    get_cluster_status_text,
    get_ticket_status_text,
)
from pcs.lib.resource_agent import STONITH_ACTION_REPLACED_BY
from pcs.lib.sbd import get_sbd_service_name


class _ServiceStatus(NamedTuple):
    service: str
    display_always: bool
    enabled: bool
    running: bool


def full_cluster_status_plaintext(
    env: LibraryEnvironment,
    hide_inactive_resources: bool = False,
    verbose: bool = False,
) -> str:
    """
    Return full cluster status as plaintext

    env -- LibraryEnvironment
    hide_inactive_resources -- if True, do not display non-running resources
    verbose -- if True, display more info
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements

    # validation
    if not env.is_cib_live and env.is_corosync_conf_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentNotConsistent(
                    [file_type_codes.CIB],
                    [file_type_codes.COROSYNC_CONF],
                )
            )
        )
    if env.is_cib_live and not env.is_corosync_conf_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentNotConsistent(
                    [file_type_codes.COROSYNC_CONF],
                    [file_type_codes.CIB],
                )
            )
        )

    # initialization
    runner = env.cmd_runner()
    report_processor = env.report_processor
    live = env.is_cib_live and env.is_corosync_conf_live
    is_sbd_running = False

    # load status, cib, corosync.conf
    status_text, warning_list = get_cluster_status_text(
        runner, hide_inactive_resources, verbose
    )
    corosync_conf = None
    # If we are live on a remote node, we have no corosync.conf.
    # TODO Use the new file framework so the path is not exposed.
    if not live or os.path.exists(settings.corosync_conf_file):
        corosync_conf = env.get_corosync_conf()
    cib = env.get_cib()
    if verbose:
        (
            ticket_status_text,
            ticket_status_stderr,
            ticket_status_retval,
        ) = get_ticket_status_text(runner)
    # get extra info if live
    if live:
        service_manager = env.service_manager
        try:
            is_sbd_running = service_manager.is_running(
                get_sbd_service_name(service_manager)
            )
        except LibraryError:
            pass
        local_services_status = _get_local_services_status(service_manager)
        if verbose and corosync_conf:
            node_name_list, node_names_report_list = get_existing_nodes_names(
                corosync_conf
            )
            report_processor.report_list(node_names_report_list)
            node_reachability = _get_node_reachability(
                env.get_node_target_factory(),
                env.get_node_communicator(),
                report_processor,
                node_name_list,
            )

    # check stonith configuration
    warning_list = list(warning_list)
    warning_list.extend(_stonith_warnings(cib, is_sbd_running))

    # put it all together
    if report_processor.has_errors:
        raise LibraryError()

    cluster_name = (
        corosync_conf.get_cluster_name()
        if corosync_conf
        else nvpair.get_value(
            "cluster_property_set", get_crm_config(cib), "cluster-name", ""
        )
    )
    parts = []
    parts.append(f"Cluster name: {cluster_name}")
    if warning_list:
        parts.extend(["", "WARNINGS:"] + warning_list + [""])
    parts.append(status_text)
    if verbose:
        parts.extend(["", "Tickets:"])
        if ticket_status_retval != 0:
            ticket_warning_parts = [
                "WARNING: Unable to get information about tickets"
            ]
            if ticket_status_stderr:
                ticket_warning_parts.extend(
                    indent(ticket_status_stderr.splitlines())
                )
            parts.extend(indent(ticket_warning_parts))
        else:
            parts.extend(indent(ticket_status_text.splitlines()))
    if live:
        if verbose and corosync_conf:
            parts.extend(["", "PCSD Status:"])
            parts.extend(
                indent(
                    _format_node_reachability(node_name_list, node_reachability)
                )
            )
        parts.extend(["", "Daemon Status:"])
        parts.extend(
            indent(_format_local_services_status(local_services_status))
        )
    return "\n".join(parts)


def _stonith_warnings(cib: _Element, is_sbd_running: bool) -> List[str]:
    warning_list = []

    is_stonith_enabled = stonith.is_stonith_enabled(get_crm_config(cib))
    (
        stonith_all,
        stonith_with_action,
        stonith_with_method_cycle,
    ) = stonith.get_misconfigured_resources(get_resources(cib))

    if is_stonith_enabled and not stonith_all and not is_sbd_running:
        warning_list.append(
            "No stonith devices and stonith-enabled is not false"
        )

    if stonith_with_action:
        warning_list.append(
            (
                "Following stonith devices have the 'action' option set, "
                "it is recommended to set {0} instead: {1}"
            ).format(
                format_list(list(STONITH_ACTION_REPLACED_BY)),
                format_list(
                    [str(x.get("id", "")) for x in stonith_with_action]
                ),
            )
        )

    if stonith_with_method_cycle:
        warning_list.append(
            "Following stonith devices have the 'method' option set "
            "to 'cycle' which is potentially dangerous, please consider using "
            "'onoff': {0}".format(
                format_list(
                    [str(x.get("id", "")) for x in stonith_with_method_cycle]
                ),
            )
        )

    return warning_list


def _get_local_services_status(
    service_manager: ServiceManagerInterface,
) -> List[_ServiceStatus]:
    service_def = [
        # (service name, display even if not enabled nor running)
        ("corosync", True),
        ("pacemaker", True),
        ("pacemaker_remote", False),
        ("pcsd", True),
        (get_sbd_service_name(service_manager), False),
    ]
    service_status_list = []
    for service, display_always in service_def:
        try:
            service_status_list.append(
                _ServiceStatus(
                    service,
                    display_always,
                    service_manager.is_enabled(service),
                    service_manager.is_running(service),
                )
            )
        except LibraryError:
            pass
    return service_status_list


def _format_local_services_status(
    service_status_list: Iterable[_ServiceStatus],
) -> List[str]:
    return [
        "{service}: {active}/{enabled}".format(
            service=status.service,
            active=("active" if status.running else "inactive"),
            enabled=("enabled" if status.enabled else "disabled"),
        )
        for status in service_status_list
        if status.display_always or status.enabled or status.running
    ]


def _get_node_reachability(
    node_target_factory: NodeTargetLibFactory,
    node_communicator: Communicator,
    report_processor: ReportProcessor,
    node_name_list: Iterable[str],
) -> Mapping[str, str]:
    # we are not interested in reports telling the user which nodes are
    # unknown since we display that info in the list of nodes
    (
        dummy_report_list,
        target_list,
    ) = node_target_factory.get_target_list_with_reports(node_name_list)
    com_cmd = CheckReachability(report_processor)
    com_cmd.set_targets(target_list)
    return run_communication(node_communicator, com_cmd)


def _format_node_reachability(
    node_name_list: Iterable[str], node_reachability: Mapping[str, str]
) -> List[str]:
    translate = {
        CheckReachability.REACHABLE: "Online",
        CheckReachability.UNAUTH: "Unable to authenticate",
        CheckReachability.UNREACHABLE: "Offline",
    }
    return [
        "{node}: {status}".format(
            node=node_name,
            status=translate[
                node_reachability.get(node_name, CheckReachability.UNAUTH)
            ],
        )
        for node_name in sorted(node_name_list)
    ]



# pylint: disable=too-many-lines,no-member
import json
import re

from functools import partial
from textwrap import dedent
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import outdent

from pcs import settings
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import cluster

QDEVICE_HOST = "qdevice.host"
CLUSTER_NAME = "myCluster"


def _corosync_options_fixture(option_list, indent_level=2):
    indent = indent_level * 4 * " "
    return "".join(
        [f"{indent}{option}: {value}\n" for option, value in option_list]
    )


def _get_two_node(nodes_num):
    if nodes_num == 2:
        return [("two_node", "1")]
    return []


def corosync_conf_fixture(
    node_list=(), quorum_options=(), qdevice_net=False, qdevice_tie_breaker=None
):
    nodes = []
    for node in node_list:
        nodes.append(
            dedent(
                """\
                node {{
            {options}    }}
            """
            ).format(options=_corosync_options_fixture(node))
        )
    device = ""
    if qdevice_net:
        if qdevice_tie_breaker:
            qd_tie_breaker = f"\n            tie_breaker: {qdevice_tie_breaker}"
        else:
            qd_tie_breaker = ""
        device = outdent(
            """
                device {{
                    model: net

                    net {{
                        host: {QDEVICE_HOST}{qd_tie_breaker}
                    }}
                }}
            """
        ).format(
            QDEVICE_HOST=QDEVICE_HOST,
            qd_tie_breaker=qd_tie_breaker,
        )
    return dedent(
        """\
        totem {{
            version: 2
            cluster_name: {cluster_name}
            transport: knet
        }}

        nodelist {{
        {nodes}}}

        quorum {{
            provider: corosync_votequorum
        {quorum}{device}}}

        logging {{
            to_logfile: yes
            logfile: {logfile}
            to_syslog: yes
        }}
        """
    ).format(
        cluster_name=CLUSTER_NAME,
        nodes="\n".join(nodes),
        quorum=_corosync_options_fixture(quorum_options, indent_level=1),
        device=device,
        logfile=settings.corosync_log_file,
    )


def corosync_node_fixture(node_id, node, addrs):
    return [(f"ring{i}_addr", addr) for i, addr in enumerate(addrs)] + [
        ("name", node),
        ("nodeid", str(node_id)),
    ]


def node_fixture(node, node_id, addr_sufix=""):
    return corosync_node_fixture(node_id, node, [f"{node}{addr_sufix}"])


class LocalConfig:
    def __init__(self, call_collection, wrap_helper, config):
        del wrap_helper, call_collection
        self.config = config
        self.expected_reports = []

    def set_expected_reports_list(self, expected_reports):
        self.expected_reports = expected_reports

    def distribute_and_reload_corosync_conf(
        self, corosync_conf_content, node_list
    ):
        local_prefix = "local.distribute_and_reload_corosync_conf."
        (
            self.config.http.corosync.set_corosync_conf(
                corosync_conf_content,
                node_labels=node_list,
                name=f"{local_prefix}http.corosync.set_corosync_conf",
            ).http.corosync.reload_corosync_conf(
                node_labels=node_list[:1],
                name=f"{local_prefix}http.corosync.reload_corosync_conf",
            )
        )
        self.expected_reports.extend(
            [fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED)]
            + [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                    node=node,
                )
                for node in node_list
            ]
            + [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_RELOADED,
                    node=node_list[0],
                )
            ]
        )

    def destroy_cluster(self, node_list):
        local_prefix = "local.destroy_cluster."
        self.config.http.host.cluster_destroy(
            node_labels=node_list,
            name=f"{local_prefix}http.host.cluster_destroy",
        )
        self.expected_reports.extend(
            [
                fixture.info(
                    report_codes.CLUSTER_DESTROY_STARTED,
                    host_name_list=node_list,
                )
            ]
            + [
                fixture.info(
                    report_codes.CLUSTER_DESTROY_SUCCESS,
                    node=node,
                )
                for node in node_list
            ]
        )


get_env_tools = partial(get_env_tools, local_extensions={"local": LocalConfig})


class SuccessMinimal(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.existing_nodes = []
        self.nodes_to_remove = []
        self.nodes_to_stay = []
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)

    def set_up(self, staying_num, removing_num):
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        sbd_installed_check = "is_sbd_installed"
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            # SBD not installed
            .services.is_installed(
                "sbd", return_value=False, name=sbd_installed_check
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    existing_corosync_nodes[:-removing_num],
                    _get_two_node(staying_num),
                ),
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )
        if len(self.nodes_to_stay) % 2 != 0:
            self.config.calls.remove(sbd_installed_check)

    def test_2_staying_1_removed(self):
        self.set_up(2, 1)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_3_staying_1_removed(self):
        self.set_up(3, 1)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_1_removed(self):
        self.set_up(4, 1)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_3_staying_2_removed(self):
        self.set_up(3, 2)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_2_removed(self):
        self.set_up(4, 2)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_3_removed(self):
        self.set_up(4, 3)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_1_staying_1_removed_forced(self):
        self.set_up(1, 1)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_1_staying_2_removed_forced(self):
        self.set_up(1, 2)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_1_staying_3_removed_forced(self):
        self.set_up(1, 3)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_2_staying_2_removed_forced(self):
        self.set_up(2, 2)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_2_staying_3_removed_forced(self):
        self.set_up(2, 3)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_3_staying_3_removed_forced(self):
        self.set_up(3, 3)
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )


class NodeNamesMissing(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)

        staying_num = 2
        removing_num = 1
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        self.original_conf = corosync_conf_fixture(
            existing_corosync_nodes,
            _get_two_node(len(self.existing_nodes)),
        )
        self.expected_conf = corosync_conf_fixture(
            existing_corosync_nodes[:-removing_num], _get_two_node(staying_num)
        )

    def test_some_node_names_missing(self):
        original_conf = re.sub(r"\s+name: node0\n", "\n", self.original_conf)
        # make sure the name was removed
        self.assertNotEqual(original_conf, self.original_conf)

        (self.config.corosync_conf.load_content(original_conf))

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(), self.nodes_to_remove
            ),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_all_node_names_missing(self):
        original_conf = re.sub(r"\s+name: .*\n", "\n", self.original_conf)
        # make sure the names were removed
        self.assertNotEqual(original_conf, self.original_conf)

        (self.config.corosync_conf.load_content(original_conf))

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(), self.nodes_to_remove
            ),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
                fixture.error(
                    report_codes.NODE_NOT_FOUND, node="node2", searched_types=[]
                ),
            ]
        )


class SuccessAtbRequired(TestCase):
    # This also tests a case when the cluster is not running on all nodes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.existing_nodes = []
        self.nodes_to_remove = []
        self.nodes_to_stay = []
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)

    def set_up(self, staying_num, removing_num):
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .services.is_installed("sbd", return_value=True)
            .services.is_enabled("sbd", return_value=True)
            .http.corosync.check_corosync_offline(
                node_labels=self.nodes_to_stay,
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .http.corosync.set_corosync_conf(
                corosync_conf_fixture(
                    existing_corosync_nodes[:-removing_num],
                    [("auto_tie_breaker", "1")],
                ),
                node_labels=self.nodes_to_stay,
            )
            # corosync is not running
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        dict(
                            label=node,
                            output=json.dumps(
                                dict(code="not_running", message="")
                            ),
                        )
                    ]
                    for node in self.nodes_to_stay
                ]
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )
        self.expected_reports.extend(
            [
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_ATB_WILL_BE_ENABLED_DUE_TO_SBD
                ),
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            ]
            + [
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
            + [fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED)]
            + [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
            + [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
        )

    def test_2_staying_1_removed(self):
        self.set_up(2, 1)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_2_staying_2_removed(self):
        self.set_up(2, 2)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_2_staying_3_removed(self):
        self.set_up(2, 3)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_1_removed(self):
        self.set_up(4, 1)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_2_removed(self):
        self.set_up(4, 2)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_3_removed(self):
        self.set_up(4, 3)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_4_removed(self):
        self.set_up(4, 4)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)

    def test_4_staying_5_removed(self):
        self.set_up(4, 5)
        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)
        self.env_assist.assert_reports(self.expected_reports)


class FailureAtbRequired(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        staying_num = 4
        removing_num = 2
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .services.is_installed("sbd", return_value=True)
            .services.is_enabled("sbd", return_value=True)
        )
        self.expected_reports.extend(
            [
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_ATB_WILL_BE_ENABLED_DUE_TO_SBD
                ),
                fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            ]
        )

    def test_failed_on_some(self):
        err_output = "an error"
        self.config.http.corosync.check_corosync_offline(
            communication_list=[
                {"label": self.nodes_to_stay[0]},
                {
                    "label": self.nodes_to_stay[1],
                    "was_connected": False,
                    "errno": 1,
                    "error_msg": err_output,
                },
                {"label": self.nodes_to_stay[2]},
                {
                    "label": self.nodes_to_stay[3],
                    "output": "not json",
                },
            ]
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.info(
                    report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                    node=node,
                )
                for node in [self.nodes_to_stay[0], self.nodes_to_stay[2]]
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_stay[1],
                    command="remote/status",
                    reason=err_output,
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.nodes_to_stay[1],
                    force_code=None,
                ),
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=self.nodes_to_stay[3],
                    force_code=None,
                ),
            ]
        )

    def test_failed_all(self):
        err_output = "an error"
        self.config.http.corosync.check_corosync_offline(
            communication_list=[
                {
                    "label": node,
                    "was_connected": False,
                    "errno": 1,
                    "error_msg": err_output,
                }
                for node in self.nodes_to_stay
            ]
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/status",
                    reason=err_output,
                )
                for node in self.nodes_to_stay
            ]
            + [
                fixture.error(
                    report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
        )


class QuorumCheck(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        staying_num = 5
        removing_num = 4
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            ).http.host.check_auth(node_labels=self.existing_nodes)
        )
        self.updated_corosync_conf_text = corosync_conf_fixture(
            existing_corosync_nodes[:-removing_num]
        )

    def test_some_nodes_not_responding(self):
        err_msg = "an error"
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=self.nodes_to_remove[0],
                            was_connected=False,
                            errno=1,
                            error_msg=err_msg,
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_remove[1],
                            response_code=400,
                            output=err_msg,
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_remove[2],
                            output="Cannot initialize CMAP service",
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_remove[3],
                        )
                    ],
                ],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
            ]
        )

    def test_all_nodes_not_responding(self):
        err_msg = "an error"
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=node,
                            was_connected=False,
                            errno=1,
                            error_msg=err_msg,
                        )
                    ]
                    for node in self.nodes_to_remove
                ],
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/get_quorum_info",
                    reason=err_msg,
                )
                for node in self.nodes_to_remove
            ]
            + [
                fixture.error(
                    report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK,
                    force_code=report_codes.FORCE,
                )
            ]
        )

    def test_all_nodes_not_responding_forced(self):
        err_msg = "an error"
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=node,
                            was_connected=False,
                            errno=1,
                            error_msg=err_msg,
                        )
                    ]
                    for node in self.nodes_to_remove
                ],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/get_quorum_info",
                    reason=err_msg,
                )
                for node in self.nodes_to_remove
            ]
            + [fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK)]
        )

    def test_all_nodes_not_running_cluster(self):
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=node,
                            output="Cannot initialize CMAP service",
                        )
                    ]
                    for node in self.nodes_to_remove
                ],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(self.expected_reports)


class FailureQuorumLoss(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        staying_num = 3
        removing_num = 3
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            ).http.host.check_auth(node_labels=self.existing_nodes)
        )
        self.updated_corosync_conf_text = corosync_conf_fixture(
            existing_corosync_nodes[:-removing_num]
        )

    def test_unable_to_get_quorum_status(self):
        err_msg = "Failure"
        self.config.http.host.get_quorum_status(
            self.existing_nodes,
            communication_list=[
                [
                    dict(
                        label=self.nodes_to_remove[0],
                        was_connected=False,
                        errno=1,
                        error_msg=err_msg,
                    )
                ],
                [
                    dict(
                        label=self.nodes_to_remove[1],
                        response_code=400,
                        output=err_msg,
                    )
                ],
                [
                    dict(
                        label=self.nodes_to_remove[2],
                        output="not parsable output",
                    )
                ],
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_GET_STATUS_ERROR,
                    reason=(
                        "Missing required section(s): 'node_list', 'quorate', "
                        "'quorum'"
                    ),
                    node=self.nodes_to_remove[2],
                ),
                fixture.error(
                    report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK,
                    force_code=report_codes.FORCE,
                ),
            ]
        )

    def test_unable_to_get_quorum_status_force(self):
        err_msg = "Failure"
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=self.nodes_to_remove[0],
                            was_connected=False,
                            errno=1,
                            error_msg=err_msg,
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_remove[1],
                            response_code=400,
                            output=err_msg,
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_remove[2],
                            output="not parsable output",
                        )
                    ],
                ],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/get_quorum_info",
                    reason=err_msg,
                ),
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_GET_STATUS_ERROR,
                    reason=(
                        "Missing required section(s): 'node_list', 'quorate', "
                        "'quorum'"
                    ),
                    node=self.nodes_to_remove[2],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK),
            ]
        )

    def test_unable_to_parse_quorum_status(self):
        self.config.http.host.get_quorum_status(
            self.existing_nodes,
            communication_list=[
                [
                    dict(
                        label=node,
                        output="not parsable output",
                    )
                ]
                for node in self.nodes_to_remove
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_GET_STATUS_ERROR,
                    reason=(
                        "Missing required section(s): 'node_list', 'quorate', "
                        "'quorum'"
                    ),
                    node=node,
                )
                for node in self.nodes_to_remove
            ]
            + [
                fixture.error(
                    report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK,
                    force_code=report_codes.FORCE,
                ),
            ]
        )

    def test_unable_to_parse_quorum_status_force(self):
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        dict(
                            label=node,
                            output="not parsable output",
                        )
                    ]
                    for node in self.nodes_to_remove
                ],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.COROSYNC_QUORUM_GET_STATUS_ERROR,
                    reason=(
                        "Missing required section(s): 'node_list', 'quorate', "
                        "'quorum'"
                    ),
                    node=node,
                )
                for node in self.nodes_to_remove
            ]
            + [fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK)]
        )

    def test_quorum_will_be_lost_force(self):
        (
            self.config.http.host.get_quorum_status(
                self.existing_nodes, node_labels=self.nodes_to_remove[:1]
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE],
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST)]
        )

    def test_quorum_will_be_lost(self):
        self.config.http.host.get_quorum_status(
            self.existing_nodes, node_labels=self.nodes_to_remove[:1]
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.COROSYNC_QUORUM_WILL_BE_LOST,
                    force_code=report_codes.FORCE,
                ),
            ]
        )


class FailureRemoveFromCib(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        self.reason = "some error"
        staying_num = 3
        removing_num = 2
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    existing_corosync_nodes[:-removing_num],
                    _get_two_node(staying_num),
                ),
                self.nodes_to_stay,
            )
        )

    def test_communication_failure(self):
        (
            self.config.http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                communication_list=[
                    dict(
                        label=self.nodes_to_stay[0],
                        was_connected=False,
                        errno=1,
                        error_msg=self.reason,
                    ),
                    dict(
                        label=self.nodes_to_stay[1],
                        response_code=400,
                        output=self.reason,
                    ),
                    dict(
                        label=self.nodes_to_stay[2],
                    ),
                ],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_stay[0],
                    command="remote/remove_nodes_from_cib",
                    reason=self.reason,
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_stay[1],
                    command="remote/remove_nodes_from_cib",
                    reason=self.reason,
                ),
            ]
        )

    def test_failure(self):
        (
            self.config.http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                communication_list=[
                    dict(
                        label=self.nodes_to_stay[0],
                        output=json.dumps(
                            dict(
                                code="failed",
                                message=self.reason,
                            )
                        ),
                    ),
                    dict(
                        label=self.nodes_to_stay[1],
                        output="not json",
                    ),
                    dict(
                        label=self.nodes_to_stay[2],
                    ),
                ],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                    node=self.nodes_to_stay[0],
                    node_list_to_remove=self.nodes_to_remove,
                    reason=self.reason,
                ),
                fixture.error(
                    report_codes.INVALID_RESPONSE_FORMAT,
                    node=self.nodes_to_stay[1],
                ),
            ]
        )

    def test_all_failed(self):
        (
            self.config.http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                communication_list=[
                    dict(
                        label=node,
                        output=json.dumps(
                            dict(
                                code="failed",
                                message=self.reason,
                            )
                        ),
                    )
                    for node in self.nodes_to_stay
                ],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                    node=node,
                    node_list_to_remove=self.nodes_to_remove,
                    reason=self.reason,
                )
                for node in self.nodes_to_stay
            ]
        )


class FailureCorosyncReload(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        self.cmd_url = "remote/reload_corosync_conf"
        self.err_msg = "An error"
        staying_num = 3
        removing_num = 2
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .http.corosync.set_corosync_conf(
                corosync_conf_fixture(
                    existing_corosync_nodes[:-removing_num],
                    _get_two_node(staying_num),
                ),
                node_labels=self.nodes_to_stay,
            )
        )
        self.expected_reports.extend(
            [fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED)]
            + [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
        )

    def test_few_failed(self):
        (
            self.config.http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        dict(
                            label=self.nodes_to_stay[0],
                            was_connected=False,
                            errno=7,
                            error_msg=self.err_msg,
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_stay[1],
                            output=json.dumps(
                                dict(
                                    code="failed",
                                    message=self.err_msg,
                                )
                            ),
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_stay[2],
                        )
                    ],
                ]
            ).http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_stay[0],
                    command=self.cmd_url,
                    reason=self.err_msg,
                ),
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_ERROR,
                    node=self.nodes_to_stay[1],
                    reason=self.err_msg,
                ),
                fixture.info(
                    report_codes.COROSYNC_CONFIG_RELOADED,
                    node=self.nodes_to_stay[2],
                ),
            ]
        )

    def test_failed_and_corosync_not_running(self):
        (
            self.config.http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        dict(
                            label=self.nodes_to_stay[0],
                            # corosync not running
                            output=json.dumps(
                                dict(code="not_running", message="")
                            ),
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_stay[1],
                            output=json.dumps(
                                dict(code="failed", message=self.err_msg)
                            ),
                        )
                    ],
                    [
                        dict(
                            label=self.nodes_to_stay[2],
                        )
                    ],
                ]
            ).http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                    node=self.nodes_to_stay[0],
                ),
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_ERROR,
                    node=self.nodes_to_stay[1],
                    reason=self.err_msg,
                ),
                fixture.info(
                    report_codes.COROSYNC_CONFIG_RELOADED,
                    node=self.nodes_to_stay[2],
                ),
            ]
        )

    def test_all_corosync_not_running(self):
        (
            self.config.http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        dict(
                            label=node,
                            # corosync not running
                            output=json.dumps(
                                dict(code="not_running", message="")
                            ),
                        )
                    ]
                    for node in self.nodes_to_stay
                ]
            ).http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
        )

    def test_all_failed(self):
        self.config.http.corosync.reload_corosync_conf(
            communication_list=[
                [
                    dict(
                        label=self.nodes_to_stay[0],
                        output=json.dumps(
                            dict(code="failed", message=self.err_msg)
                        ),
                    )
                ],
                [
                    dict(
                        label=self.nodes_to_stay[1],
                        output="not a json",
                    )
                ],
                [
                    dict(
                        label=self.nodes_to_stay[2],
                        response_code=400,
                        output=self.err_msg,
                    )
                ],
            ]
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_RELOAD_ERROR,
                    node=self.nodes_to_stay[0],
                    reason=self.err_msg,
                ),
                fixture.warn(
                    report_codes.INVALID_RESPONSE_FORMAT,
                    node=self.nodes_to_stay[1],
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_stay[2],
                    command=self.cmd_url,
                    reason=self.err_msg,
                ),
            ]
            + [
                fixture.error(
                    report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE,
                )
            ]
        )


class FailureCorosyncConfDistribution(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        staying_num = 3
        removing_num = 2
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
        )
        self.expected_reports.extend(
            [fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED)]
        )
        self.updated_corosync_conf_text = corosync_conf_fixture(
            existing_corosync_nodes[:-removing_num]
        )

    def test_failed_on_some(self):
        err_output = "an error"
        self.config.http.corosync.set_corosync_conf(
            self.updated_corosync_conf_text,
            communication_list=[
                {"label": self.nodes_to_stay[0]},
                {
                    "label": self.nodes_to_stay[1],
                    "was_connected": False,
                    "errno": 1,
                    "error_msg": err_output,
                },
                {"label": self.nodes_to_stay[2]},
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.info(
                    report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE, node=node
                )
                for node in [self.nodes_to_stay[0], self.nodes_to_stay[2]]
            ]
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_stay[1],
                    command="remote/set_corosync_conf",
                    reason=err_output,
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                    node=self.nodes_to_stay[1],
                ),
            ]
        )

    def test_failed_all(self):
        err_output = "an error"
        self.config.http.corosync.set_corosync_conf(
            self.updated_corosync_conf_text,
            communication_list=[
                {
                    "label": node,
                    "was_connected": False,
                    "errno": 1,
                    "error_msg": err_output,
                }
                for node in self.nodes_to_stay
            ],
        )

        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
            )
        )

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/set_corosync_conf",
                    reason=err_output,
                )
                for node in self.nodes_to_stay
            ]
            + [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                    node=node,
                )
                for node in self.nodes_to_stay
            ]
        )


class FailureClusterDestroy(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        staying_num = 5
        removing_num = 3
        self.existing_nodes = [
            f"node{i}" for i in range(staying_num + removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-removing_num:]
        self.nodes_to_stay = self.existing_nodes[:-removing_num]
        existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
            .http.host.check_auth(node_labels=self.existing_nodes)
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
        )
        self.updated_corosync_conf_text = corosync_conf_fixture(
            existing_corosync_nodes[:-removing_num]
        )
        self.expected_reports.extend(
            [
                fixture.info(
                    report_codes.CLUSTER_DESTROY_STARTED,
                    host_name_list=self.nodes_to_remove,
                )
            ]
        )

    def test_failed_on_some(self):
        err_output = "an error"
        (
            self.config.http.host.cluster_destroy(
                communication_list=[
                    {"label": self.nodes_to_remove[0]},
                    {
                        "label": self.nodes_to_remove[1],
                        "was_connected": False,
                        "errno": 1,
                        "error_msg": err_output,
                    },
                    {"label": self.nodes_to_remove[2]},
                ]
            )
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.info(report_codes.CLUSTER_DESTROY_SUCCESS, node=node)
                for node in [self.nodes_to_remove[0], self.nodes_to_remove[2]]
            ]
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[1],
                    command="remote/cluster_destroy",
                    reason=err_output,
                ),
                fixture.warn(
                    report_codes.NODES_TO_REMOVE_UNREACHABLE,
                    node_list=[self.nodes_to_remove[1]],
                ),
            ]
        )

    def test_failed_all(self):
        err_output = "an error"
        (
            self.config.http.host.cluster_destroy(
                communication_list=[
                    {
                        "label": node,
                        "was_connected": False,
                        "errno": 1,
                        "error_msg": err_output,
                    }
                    for node in self.nodes_to_remove
                ]
            )
            .local.distribute_and_reload_corosync_conf(
                self.updated_corosync_conf_text,
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )

        cluster.remove_nodes(self.env_assist.get_env(), self.nodes_to_remove)

        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=node,
                    command="remote/cluster_destroy",
                    reason=err_output,
                )
                for node in self.nodes_to_remove
            ]
            + [
                fixture.warn(
                    report_codes.NODES_TO_REMOVE_UNREACHABLE,
                    node_list=self.nodes_to_remove,
                )
            ]
        )


class FailureValidationCorosync(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.existing_nodes = [f"node{i}" for i in range(3)]
        self.existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.config.env.set_known_nodes(self.existing_nodes)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    self.existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                    qdevice_net=True,
                    qdevice_tie_breaker=1,
                )
            )
        )

    def test_corosync_validation(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(), self.existing_nodes + ["nodeX"]
            ),
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND, node="nodeX", searched_types=[]
                ),
                fixture.error(report_codes.CANNOT_REMOVE_ALL_CLUSTER_NODES),
                fixture.error(
                    report_codes.NODE_USED_AS_TIE_BREAKER,
                    node="node0",
                    node_id="1",
                ),
            ]
        )


class OfflineNodes(TestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.removing_num = 2
        self.staying_num = 2
        self.existing_nodes = [
            f"node{i}" for i in range(self.staying_num + self.removing_num)
        ]
        self.nodes_to_remove = self.existing_nodes[-self.removing_num :]
        self.nodes_to_stay = self.existing_nodes[: -self.removing_num]
        self.existing_corosync_nodes = [
            node_fixture(node, i)
            for i, node in enumerate(self.existing_nodes, 1)
        ]
        self.expected_reports = []
        self.config.local.set_expected_reports_list(self.expected_reports)
        (
            self.config.corosync_conf.load_content(
                corosync_conf_fixture(
                    self.existing_corosync_nodes,
                    _get_two_node(len(self.existing_nodes)),
                )
            )
        )

    def test_all_remaining_offline(self):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes
            ).http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_stay[1],
                        "response_code": 400,
                        "output": "an error",
                    },
                    {
                        "label": self.nodes_to_remove[0],
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(), self.nodes_to_remove
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    node=self.nodes_to_stay[0],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_stay[1],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.error(
                    report_codes.UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE
                ),
            ]
        )

    def test_all_remaining_offline_forced(self):
        (
            self.config.env.set_known_nodes(
                self.existing_nodes
            ).http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_stay[1],
                        "response_code": 400,
                        "output": "an error",
                    },
                    {
                        "label": self.nodes_to_remove[0],
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.SKIP_OFFLINE_NODES],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.OMITTING_NODE,
                    node=self.nodes_to_stay[0],
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_stay[1],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.error(
                    report_codes.UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE
                ),
            ]
        )

    def test_some_remaining_offline(self):
        (
            self.config.env.set_known_nodes(self.existing_nodes)
            .http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_stay[1],
                    },
                    {
                        "label": self.nodes_to_remove[0],
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
            # SBD not installed
            .services.is_installed(
                "sbd", return_value=False, name="is_sbd_installed"
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.FORCE],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    node=self.nodes_to_stay[0],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_some_remaining_offline_forced(self):
        (
            self.config.env.set_known_nodes(self.existing_nodes)
            .http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_stay[1],
                    },
                    {
                        "label": self.nodes_to_remove[0],
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
            # SBD not installed
            .services.is_installed(
                "sbd", return_value=False, name="is_sbd_installed"
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes[: -self.removing_num],
                    _get_two_node(self.staying_num),
                ),
                [self.nodes_to_stay[1]],
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=[self.nodes_to_stay[1]],
            )
        )
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.OMITTING_NODE,
                    node=self.nodes_to_stay[0],
                ),
                fixture.warn(
                    report_codes.UNABLE_TO_CONNECT_TO_ALL_REMAINING_NODE,
                    node_list=[self.nodes_to_stay[0]],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_removed_offline(self):
        (
            self.config.env.set_known_nodes(self.existing_nodes)
            .http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                    },
                    {
                        "label": self.nodes_to_stay[1],
                    },
                    {
                        "label": self.nodes_to_remove[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_remove[1],
                        "response_code": 400,
                        "output": "an error",
                    },
                ]
            )
            # SBD not installed
            .services.is_installed(
                "sbd", return_value=False, name="is_sbd_installed"
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        {
                            "label": self.nodes_to_remove[0],
                            "was_connected": False,
                            "errno": 7,
                            "error_msg": "an error",
                        }
                    ],
                    [
                        {
                            "label": self.nodes_to_remove[1],
                            "response_code": 400,
                            "output": "an error",
                        }
                    ],
                ],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.FORCE],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    node=self.nodes_to_remove[0],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/get_quorum_info",
                    reason="an error",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/get_quorum_info",
                    reason="an error",
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK),
            ]
        )

    def test_removed_offline_unforceable(self):
        (
            self.config.env.set_known_nodes(self.existing_nodes)
            .http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                    },
                    {
                        "label": self.nodes_to_stay[1],
                    },
                    {
                        "label": self.nodes_to_remove[0],
                    },
                    {
                        "label": self.nodes_to_remove[1],
                        "response_code": 400,
                        "output": "an error",
                    },
                ]
            )
            # SBD not installed
            .services.is_installed(
                "sbd", return_value=False, name="is_sbd_installed"
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[
                    report_codes.FORCE,
                    report_codes.SKIP_OFFLINE_NODES,
                ],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.nodes_to_remove[1],
                    command="remote/check_auth",
                    reason="an error",
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_removed_offline_forced(self):
        (
            self.config.env.set_known_nodes(self.existing_nodes)
            .http.host.check_auth(
                communication_list=[
                    {
                        "label": self.nodes_to_stay[0],
                    },
                    {
                        "label": self.nodes_to_stay[1],
                    },
                    {
                        "label": self.nodes_to_remove[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                communication_list=[
                    [
                        {
                            "label": self.nodes_to_remove[0],
                            "was_connected": False,
                            "errno": 7,
                            "error_msg": "an error",
                        }
                    ],
                    [
                        {
                            "label": self.nodes_to_remove[1],
                        }
                    ],
                ],
            )
            .http.host.cluster_destroy(
                communication_list=[
                    {
                        "label": self.nodes_to_remove[0],
                        "was_connected": False,
                        "errno": 7,
                        "error_msg": "an error",
                    },
                    {
                        "label": self.nodes_to_remove[1],
                    },
                ]
            )
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes[: -self.removing_num],
                    _get_two_node(self.staying_num),
                ),
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=self.nodes_to_stay,
            )
        )
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.OMITTING_NODE,
                    node=self.nodes_to_remove[0],
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/get_quorum_info",
                    reason="an error",
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
                fixture.info(
                    report_codes.CLUSTER_DESTROY_STARTED,
                    host_name_list=self.nodes_to_remove,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    node=self.nodes_to_remove[0],
                    command="remote/cluster_destroy",
                    reason="an error",
                ),
                fixture.info(
                    report_codes.CLUSTER_DESTROY_SUCCESS,
                    node=self.nodes_to_remove[1],
                ),
                fixture.warn(
                    report_codes.NODES_TO_REMOVE_UNREACHABLE,
                    node_list=[self.nodes_to_remove[0]],
                ),
            ]
        )

    def test_all_remaining_unknown(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_remove
            ).http.host.check_auth(node_labels=self.nodes_to_remove)
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(), self.nodes_to_remove
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=self.nodes_to_stay,
                ),
                fixture.error(
                    report_codes.UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE
                ),
            ]
        )

    def test_all_remaining_unknown_forced(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_remove
            ).http.host.check_auth(node_labels=self.nodes_to_remove)
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.SKIP_OFFLINE_NODES],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND, host_list=self.nodes_to_stay
                ),
                fixture.error(
                    report_codes.UNABLE_TO_CONNECT_TO_ANY_REMAINING_NODE
                ),
            ]
        )

    def test_some_remaining_unknown(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_remove + [self.nodes_to_stay[0]]
            )
            .http.host.check_auth(
                node_labels=([self.nodes_to_stay[0]] + self.nodes_to_remove)
            )
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.FORCE],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=[self.nodes_to_stay[1]],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_some_remaining_unknown_forced(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_remove + [self.nodes_to_stay[0]]
            )
            .http.host.check_auth(
                node_labels=([self.nodes_to_stay[0]] + self.nodes_to_remove)
            )
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster(self.nodes_to_remove)
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes[: -self.removing_num],
                    _get_two_node(self.staying_num),
                ),
                [self.nodes_to_stay[0]],
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove,
                node_labels=[self.nodes_to_stay[0]],
            )
        )
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND,
                    host_list=[self.nodes_to_stay[1]],
                ),
                fixture.warn(
                    report_codes.UNABLE_TO_CONNECT_TO_ALL_REMAINING_NODE,
                    node_list=[self.nodes_to_stay[1]],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_all_removed_unknown(self):
        (
            self.config.env.set_known_nodes(self.nodes_to_stay)
            .http.host.check_auth(node_labels=(self.nodes_to_stay))
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=[],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.FORCE],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=self.nodes_to_remove,
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK),
            ]
        )

    def test_all_removed_unknown_forced(self):
        (
            self.config.env.set_known_nodes(self.nodes_to_stay)
            .http.host.check_auth(node_labels=(self.nodes_to_stay))
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=[],
            )
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes[: -self.removing_num],
                    _get_two_node(self.staying_num),
                ),
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove, node_labels=self.nodes_to_stay
            )
        )
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND, host_list=self.nodes_to_remove
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_LOSS_UNABLE_TO_CHECK),
                fixture.warn(
                    report_codes.NODES_TO_REMOVE_UNREACHABLE,
                    node_list=self.nodes_to_remove,
                ),
            ]
        )

    def test_some_removed_unknown(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_stay + [self.nodes_to_remove[1]]
            )
            .http.host.check_auth(
                node_labels=(self.nodes_to_stay + [self.nodes_to_remove[1]])
            )
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=[self.nodes_to_remove[1]],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.remove_nodes(
                self.env_assist.get_env(),
                self.nodes_to_remove,
                force_flags=[report_codes.FORCE],
            ),
            [],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.error(
                    report_codes.HOST_NOT_FOUND,
                    force_code=report_codes.SKIP_OFFLINE_NODES,
                    host_list=[self.nodes_to_remove[0]],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
            ]
        )

    def test_some_removed_unknown_forced(self):
        (
            self.config.env.set_known_nodes(
                self.nodes_to_stay + [self.nodes_to_remove[0]]
            )
            .http.host.check_auth(
                node_labels=(self.nodes_to_stay + [self.nodes_to_remove[0]])
            )
            # SBD not installed
            .services.is_installed(
                "sbd",
                return_value=False,
                name="services.is_installed.sbd.False",
            )
            .http.host.get_quorum_status(
                self.existing_nodes,
                node_labels=self.nodes_to_remove[:1],
            )
            .local.destroy_cluster([self.nodes_to_remove[0]])
            .local.distribute_and_reload_corosync_conf(
                corosync_conf_fixture(
                    self.existing_corosync_nodes[: -self.removing_num],
                    _get_two_node(self.staying_num),
                ),
                self.nodes_to_stay,
            )
            .http.pcmk.remove_nodes_from_cib(
                self.nodes_to_remove, node_labels=self.nodes_to_stay
            )
        )
        cluster.remove_nodes(
            self.env_assist.get_env(),
            self.nodes_to_remove,
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        self.env_assist.assert_reports(
            self.expected_reports
            + [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND,
                    host_list=[self.nodes_to_remove[1]],
                ),
                fixture.warn(report_codes.COROSYNC_QUORUM_WILL_BE_LOST),
                fixture.warn(
                    report_codes.NODES_TO_REMOVE_UNREACHABLE,
                    node_list=[self.nodes_to_remove[1]],
                ),
            ]
        )



import os.path
from typing import (
    List,
    Optional,
)

from .. import errors
from ..interfaces import (
    ExecutorInterface,
    ServiceManagerInterface,
)


class SysVInitRhelDriver(ServiceManagerInterface):
    def __init__(
        self, executor: ExecutorInterface, service_bin: str, chkconfig_bin: str
    ):
        """
        executor -- external commands used by this class are executed using
            this object
        service_bin -- path to an executable used for starting and stopping
            services and to check if a service is running
        chkconfig_bin -- path to an executable used for enabling, disabling and
            listing available service and to check if service is enabled
        """
        self._executor = executor
        self._service_bin = service_bin
        self._chkconfig_bin = chkconfig_bin
        self._available_services: List[str] = []

    def start(self, service: str, instance: Optional[str] = None) -> None:
        result = self._executor.run([self._service_bin, service, "start"])
        if result.retval != 0:
            raise errors.StartServiceError(service, result.joined_output)

    def stop(self, service: str, instance: Optional[str] = None) -> None:
        result = self._executor.run([self._service_bin, service, "stop"])
        if result.retval != 0:
            raise errors.StopServiceError(service, result.joined_output)

    def enable(self, service: str, instance: Optional[str] = None) -> None:
        result = self._executor.run([self._chkconfig_bin, service, "on"])
        if result.retval != 0:
            raise errors.EnableServiceError(service, result.joined_output)

    def disable(self, service: str, instance: Optional[str] = None) -> None:
        if not self.is_installed(service):
            return
        result = self._executor.run([self._chkconfig_bin, service, "off"])
        if result.retval != 0:
            raise errors.DisableServiceError(service, result.joined_output)

    def is_enabled(self, service: str, instance: Optional[str] = None) -> bool:
        return self._executor.run([self._chkconfig_bin, service]).retval == 0

    def is_running(self, service: str, instance: Optional[str] = None) -> bool:
        return (
            self._executor.run([self._service_bin, service, "status"]).retval
            == 0
        )

    def is_installed(self, service: str) -> bool:
        return service in self.get_available_services()

    def get_available_services(self) -> List[str]:
        if not self._available_services:
            self._available_services = self._get_available_services()
        return self._available_services

    def _get_available_services(self) -> List[str]:
        result = self._executor.run([self._chkconfig_bin])
        if result.retval != 0:
            return []

        service_list = []
        for service in result.stdout.splitlines():
            service = service.split(" ", 1)[0]
            if service:
                service_list.append(service)
        return service_list

    def is_current_system_supported(self) -> bool:
        return all(
            os.path.isfile(binary)
            for binary in (self._service_bin, self._chkconfig_bin)
        )



from unittest import TestCase

from pcs.common.reports.constraints import common


class OptionsTest(TestCase):
    def test_get_console_options_from_lib_options(self):
        self.assertEqual(
            ["a=b", "c=d"],
            common.options({"c": "d", "a": "b", "id": "some_id"}),
        )


class IdFromOptionsTest(TestCase):
    def test_get_id_from_options(self):
        self.assertEqual(
            "(id:some_id)",
            common.id_from_options({"c": "d", "a": "b", "id": "some_id"}),
        )


class PrepareOptionsTest(TestCase):
    def test_prepare_options_with_id(self):
        self.assertEqual(
            ["a=b", "c=d", "(id:some_id)"],
            common.prepare_options({"c": "d", "a": "b", "id": "some_id"}),
        )

    def test_prepare_options_without_id(self):
        self.assertEqual(
            ["a=b", "c=d"],
            common.prepare_options(
                {"c": "d", "a": "b", "id": "some_id"}, with_id=False
            ),
        )


class ResourceSetsTest(TestCase):
    def test_prepare_resource_sets_without_id(self):
        self.assertEqual(
            ["set", "a", "b", "c=d", "e=f", "set", "g", "h", "i=j", "k=l"],
            common.resource_sets(
                [
                    {
                        "ids": ["a", "b"],
                        "options": {"c": "d", "e": "f", "id": "some_id"},
                    },
                    {
                        "ids": ["g", "h"],
                        "options": {"i": "j", "k": "l", "id": "some_id_2"},
                    },
                ],
                with_id=False,
            ),
        )

    def test_prepare_resource_sets_with_id(self):
        self.assertEqual(
            [
                "set",
                "a",
                "b",
                "c=d",
                "e=f",
                "(id:some_id)",
                "set",
                "g",
                "h",
                "i=j",
                "k=l",
                "(id:some_id_2)",
            ],
            common.resource_sets(
                [
                    {
                        "ids": ["a", "b"],
                        "options": {"c": "d", "e": "f", "id": "some_id"},
                    },
                    {
                        "ids": ["g", "h"],
                        "options": {"i": "j", "k": "l", "id": "some_id_2"},
                    },
                ]
            ),
        )
